
#ifndef KDB_EXPORT_H
#define KDB_EXPORT_H

#ifdef KDB_STATIC_DEFINE
#  define KDB_EXPORT
#  define KDB_NO_EXPORT
#else
#  ifndef KDB_EXPORT
#    ifdef KDb_EXPORTS
        /* We are building this library */
#      define KDB_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define KDB_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef KDB_NO_EXPORT
#    define KDB_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef KDB_DEPRECATED
#  define KDB_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef KDB_DEPRECATED_EXPORT
#  define KDB_DEPRECATED_EXPORT KDB_EXPORT KDB_DEPRECATED
#endif

#ifndef KDB_DEPRECATED_NO_EXPORT
#  define KDB_DEPRECATED_NO_EXPORT KDB_NO_EXPORT KDB_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef KDB_NO_DEPRECATED
#    define KDB_NO_DEPRECATED
#  endif
#endif

#endif /* KDB_EXPORT_H */
